# -*- coding: utf-8 -*-
"""
Created on Mon Aug 27 17:08:42 2018

@author: tb267
"""



class DataSet():
    def __init__(self,timedata=None,freqdata=None,tfdata=None,sonodata=None,metadata=None):
        self.timedata = timedata
        self.freqdata = freqdata
        self.tfdata   = tfdata
        self.sonodata = sonodata
        self.metadata = metadata
        
    def __repr__(self):
        text = '<DataSet class containing: '
        if repr(self.timedata) != 'None':
            text += repr(self.timedata)
        if repr(self.freqdata) != 'None':
            text += repr(self.freqdata)
        if repr(self.tfdata) != 'None':
            text += repr(self.tfdata)
        if repr(self.sonodata) != 'None':
            text += repr(self.sonodata)
        if repr(self.metadata) != 'None':
            text += repr(self.metadata)
        
        text += '>'
        return text
        
        
class TimeData():
    def __init__(self,time_axis,time_data,settings):
        self.time_axis = time_axis
        self.time_data = time_data  
        self.settings = settings
        
    def __repr__(self):
        return "<TimeData>"

        
class FreqData():
    def __init__(self,freq_axis,freq_data,settings):
        self.freq_axis = freq_axis
        self.freq_data = freq_data
        self.settings = settings
        
    def __repr__(self):
        return "<FreqData>"
        
class TfData():
    def __init__(self,freq_axis,tf_data,settings):
        self.freq_axis = freq_axis
        self.tf_data = tf_data
        self.settings = settings
        
    def __repr__(self):
        return "<TfData>"
        
class SonoData():
    def __init__(self,time_axis,freq_axis,sono_data,settings):
        self.time_axis = time_axis
        self.freq_axis = freq_axis
        self.sono_data = sono_data
        self.settings = settings
        
    def __repr__(self):
        return "<SonoData>"
        
class MetaData():
    def __init__(self, timestamp=None, timestring=None, units=None, channel_cal_factors=None, tf_cal_factors = None):
        self.timestamp = timestamp
        self.timestring = timestring
        self.units = units
        self.channel_cal_factors = None
        self.tf_cal_factors = None
        
    def __repr__(self):
        return "<MetaData>"


# -*- coding: utf-8 -*-
"""
Created on Sun Sep  2 20:16:26 2018

@author: tb267
"""
import sys

from . import settings
from . import file
from . import logdata
from . import plotting

import numpy as np
import scipy.signal as sp
import time
import datetime
import copy

def convert_to_frequency(time_data,time_range=None,window=False):
    '''
    Args:
        timeData (class): time series data
        time_range: 2x1 numpy array to specify data segment to use
        window (bool): apply blackman filter to data before fft or not
    '''
    
    if time_range == None:
        ### use all data
        time_range = time_data.time_axis[[0,-1]]
        
    elif time_range.__class__.__name__ == 'PlotData':
        time_range=time_range.ax.get_xbound()
        
    settings = copy.copy(time_data.settings)
    settings.window = window
    settings.time_range = time_range

    
    s1 = time_data.time_axis >= time_range[0]
    s2 = time_data.time_axis <= time_range[1]
    selection = s1 & s2
    data_selected = time_data.time_data[selection,:]
    N = len(data_selected[:,0])
    if window == True:
        data_selected = np.blackman(N)
        
    fdata = np.fft.rfft(data_selected,axis=0)
    faxis = np.fft.rfftfreq(N,1/time_data.settings.fs)
    
    freq_data = logdata.FreqData(faxis,fdata,settings)
    
    return freq_data


from .settings import MySettings
from .file import load_data, save_data
from .oscilloscope import Oscilloscope, Recorder
from .logdata import DataSet, TimeData, FreqData, TfData, SonoData, MetaData
from .plotting import PlotData
from .analysis import convert_to_frequency





# -*- coding: utf-8 -*-
"""
Created on Fri Aug  3 11:27:29 2018

@authors: ae407, tb267
"""      
import sys

from . import settings
from . import file
from . import logdata
from . import plotting

import pyaudio
import numpy as np
import pyqtgraph as pg
from pyqtgraph.Qt import QtGui, QtCore
import matplotlib
import matplotlib.pyplot as plt
import time
import datetime





class Oscilloscope():
    
    app = None
    win = None
    timer = QtCore.QTimer()

    def __init__(self,settings):
        '''Creates an Oscilloscope
        Args:
            settings: An object of the class mySettings
        '''

        
        self.settings = settings
            
        self.rec = Recorder(settings)
        self.rec.init_pyaudio(settings)
#        self.audio_stream = audio_stream
#        self.audio = audio    
        
        if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
            
#            self.timer = QtCore.QTimer()
            self.create_figure()
            
            
            Oscilloscope.win.sigKeyPress.connect(self.keyPressed)
            
            
            Oscilloscope.timer.timeout.connect(self.update) # update figure and buffer
            Oscilloscope.timer.start(0)
                
            Oscilloscope.app.instance().exec_()
            
            
    def create_figure(self):
        '''
        Creates a figure which is an object of the class KeyPressWindow.
        
        '''
        pg.setConfigOption('background', 'w')
        if Oscilloscope.app == None:
            Oscilloscope.app = QtGui.QApplication([])
            
            
#        if Oscilloscope.win == None:
        Oscilloscope.win = KeyPressWindow(self.rec)
#        Oscilloscope.win.setWindowIcon(QtGui.QIcon('icon.png'))
        
        
        window_geometry = Oscilloscope.win.geometry()
        

        
        Oscilloscope.win.setGeometry(100,100,800,600)
        
        Oscilloscope.win.showMinimized()
        Oscilloscope.win.showNormal()


        Oscilloscope.win.setWindowTitle("Oscilloscope (to save to new filename press 's', to autosave press 'space')")
        self.view_time = self.settings.init_view_time
        self.view_freq = self.settings.init_view_freq
        self.view_levels = self.settings.init_view_levels
        
        self.toggle_view()
        
        self.data_saved_counter = 0 #  to indicate not yet saved file
           
            

    def toggle_view(self):
        '''
        Switches between views, triggered by keypress
        '''
        Oscilloscope.win.clear()
        
        
        if self.view_time == True:
            self.time_plot()
            
        if self.view_freq == True:
            self.freq_plot()
            
        if self.view_levels == True:
            self.levels_plot()
        
        
            
            
    
        
    
    def time_plot(self):
        # create a plot for the time domain
        self.view_time = True
        Oscilloscope.win.nextRow()
        self.osc_time_line = Oscilloscope.win.addPlot(title="Time Domain (toggle with 'T')")
        if self.settings.channels==1:
            self.osc_time_line.enableAutoRange()
        else:
            self.osc_time_line.setYRange(-1,self.settings.channels)
        self.osc_time_line.setXRange(self.rec.osc_time_axis[0],self.rec.osc_time_axis[-1])
        self.osc_time_line.showGrid(True, True)
        self.osc_time_line.addLegend()
        self.osc_time_line.setLabel('left','Normalised Amplitude')
        self.osc_time_line.setLabel('bottom','Time (s)')  
        
        ax=self.osc_time_line.getAxis('left')
        ax.setTickSpacing(1,1)
        
        self.osc_time_lineset={}
        for i in range(self.settings.channels):
            pen_ = pg.mkPen(color=settings.set_plot_colours(self.settings.channels)[i,:])
            self.osc_time_lineset[i]=self.osc_time_line.plot(pen=pen_, name='Channel '+str(i))
        
#        Oscilloscope.win.FillBetweenItem(curve1=osc_time_lineset[0], curve2=osc_time_lineset[1])
        
    def freq_plot(self):
        # create a plot for the frequency domain
        self.view_freq = True
        Oscilloscope.win.nextRow()
        self.osc_freq_line = Oscilloscope.win.addPlot(title="Frequency Domain (toggle with 'F')") 
        self.osc_freq_line.enableAutoRange()
        self.osc_freq_line.setXRange(self.rec.osc_freq_axis[0],self.rec.osc_freq_axis[-1])
        self.osc_freq_line.showGrid(True, True)
        self.osc_freq_line.addLegend()
        self.osc_freq_line.setLabel('left','Power Spectrum (dB)')
        self.osc_freq_line.setLabel('bottom','Frequency (Hz)')
        
        self.osc_freq_lineset={}
        for i in range(self.settings.channels):
            pen_ = pg.mkPen(color=settings.set_plot_colours(self.settings.channels)[i,:])
            self.osc_freq_lineset[i]=self.osc_freq_line.plot(pen=pen_, name='Channel'+str(i))
            
    def levels_plot(self):
        # create a plot for the frequency domain
        self.view_levels = True
        Oscilloscope.win.nextRow()
        
        self.osc_levels_line = Oscilloscope.win.addPlot(title="Channel Levels (toggle with 'L')")
        self.osc_levels_line.setYRange(0,1)
        self.osc_levels_line.setXRange(-0.5,self.settings.channels-0.5)
        self.osc_levels_line.showGrid(False,True)
        self.osc_levels_line.setLabel('left','Normalised Amplitude')
        self.osc_levels_line.setLabel('bottom','Channel Index')
        
        
        ax=self.osc_levels_line.getAxis('bottom')
        ax.setTickSpacing(1,1)    
#        ax.showLabel(show=True)
#        self.osc_levels_line.setTicks(np.arange(self.settings.channels))
        self.osc_levels_lineset={}
        for i in range(self.settings.channels):
            pen_ = pg.mkPen(color=settings.set_plot_colours(self.settings.channels)[i,:],width=3)
            pen_peak = pg.mkPen(color=settings.set_plot_colours(self.settings.channels)[i,:],width=3)
            self.osc_levels_lineset[i]=self.osc_levels_line.plot(pen=pen_, name='vertical')
            self.osc_levels_lineset[self.settings.channels+i]=self.osc_levels_line.plot(pen=pen_, name='top')
            self.osc_levels_lineset[2*self.settings.channels+i]=self.osc_levels_line.plot(pen=pen_peak, name='peak hold')    
#            self.osc_levels_lineset[3]=self.osc_levels_line.plot(pen=pen_, name='Channel')
        
        self.osc_levels_peak_hold = np.zeros(self.settings.channels)
        self.time_last_changed = np.zeros(self.settings.channels)
        
    
    def update(self):
        '''
        Updates plots with incoming data from __call__. 
        Called with a 0s interval by QTimer.
        
        '''
        if self.view_levels == True:
            self.osc_levels_rms = np.sqrt(np.mean(self.rec.osc_time_data**2,axis=0))
            self.osc_levels_max = np.max(np.abs(self.rec.osc_time_data),axis=0)
            changed_indices = self.osc_levels_peak_hold < self.osc_levels_max
            self.time_last_changed[changed_indices] = time.time()
            self.osc_levels_peak_hold = np.maximum(self.osc_levels_peak_hold,self.osc_levels_max)
            self.osc_levels_peak_hold[time.time()-self.time_last_changed>2] = 0
            
                
        for i in range(self.settings.channels):
            offset=i
            if self.view_time == True:
                self.osc_time_lineset[i].setData(self.rec.osc_time_axis,self.rec.osc_time_data[:,i] + offset)
            
            if self.view_freq == True:
                # calculate the FFT
                self.rec.osc_time_data_windowed[:,i] = self.rec.osc_time_data[:,i] * np.blackman(np.shape(self.rec.osc_time_data)[0])
                self.rec.osc_freq_data[:,i] = 20 * np.log10(np.abs(np.fft.rfft(self.rec.osc_time_data_windowed[:,i]))/len(self.rec.osc_time_data_windowed[:,i]))
                self.osc_freq_lineset[i].setData(self.rec.osc_freq_axis,self.rec.osc_freq_data[:,i])
                
            if self.view_levels == True:
                self.osc_levels_lineset[i].setData([i,i],[0,self.osc_levels_max[i]])
                self.osc_levels_lineset[self.settings.channels+i].setData([i-0.3,i+0.3],self.osc_levels_max[i]*np.ones(2))
                
                if self.osc_levels_peak_hold[i] > 0.98:
                    pen_peak = pg.mkPen(color=settings.set_plot_colours(self.settings.channels)[i,:],width=10)
                else:
                    pen_peak = pg.mkPen(color=settings.set_plot_colours(self.settings.channels)[i,:],width=3)
#                self.osc_levels_lineset[2*self.settings.channels+i]=self.osc_levels_line.plot(pen=pen_peak, name='peak hold')
                self.osc_levels_lineset[2*self.settings.channels+i].setData([i-0.3,i+0.3],self.osc_levels_peak_hold[i]*np.ones(2),pen=pen_peak)
#                self.osc_levels_lineset[3].setData(np.arange(2),np.ones(2))
                
                
            #updates for the stored
            self.rec.stored_time_data_windowed[:,i] = self.rec.stored_time_data[:,i] * np.blackman(np.shape(self.rec.stored_time_data)[0])
            self.rec.stored_freq_data[:,i] = 20 * np.log10(np.abs(np.fft.rfft(self.rec.stored_time_data_windowed[:,i]))/len(self.rec.stored_time_data_windowed[:,i]))
    

        
            
            
    #KeyPressed function within osciolloscpe since can only take one argument        
    def keyPressed(self, evt):
        '''
        Upon a Space Bar press, makes a copy of data from the past stored_time seconds,plots it in Bokeh and gives the user an option to save it.
        '''
        
        if evt.key() == QtCore.Qt.Key_T:

            if self.view_freq != False or self.view_levels != False:
                print('toggled time domain view')
                self.view_time = not self.view_time
                self.toggle_view()
            else:
                print('toggling all views off is prevented')
            
            
        if evt.key() == QtCore.Qt.Key_F:

            if self.view_time != False or self.view_levels != False:
                print('toggled frequency domain view')
                self.view_freq = not self.view_freq
                self.toggle_view()
            else:
                print('toggling all views off is prevented')
            
            
        if evt.key() == QtCore.Qt.Key_L:
            if self.view_time != False or self.view_freq != False:
                print('toggled levels view')
                self.view_levels = not self.view_levels
                self.toggle_view()
            else:
                print('toggling all views off is prevented')
        
        
        if evt.key() == QtCore.Qt.Key_Space or evt.key() == QtCore.Qt.Key_S:
            
            stored_time_data_copy=np.copy(self.rec.stored_time_data)
            t = datetime.datetime.now()
            timestring = '_'+str(t.year)+'_'+str(t.month)+'_'+str(t.day)+'_at_'+str(t.hour)+'_'+str(t.minute)+'_'+str(t.second)
            
            
            print("key press trigger: saving data to file in working directory")
            
            ### make into dataset
            
            fs=self.settings.fs
            n_samp=len(stored_time_data_copy[:,0])
            dt=1/fs
            t_samp=n_samp*dt
            t_axis= np.arange(0,t_samp,dt)
#            f_axis=np.fft.rfftfreq(len(t_axis),1/fs)
#            freq_data=np.fft.rfft(stored_time_data_copy,axis=0)
            
            
            t_data = logdata.TimeData(t_axis,stored_time_data_copy,self.settings)
            metadata = logdata.MetaData(timestamp=t,timestring=timestring)
            dataset = logdata.DataSet(timedata=t_data, metadata=metadata)
            
            #plotting.plotdata(dataset)
            
            
            if evt.key() == QtCore.Qt.Key_S:
                self.data_saved_counter = 0
            
            
            if self.data_saved_counter == 0:
                self.last_filename = file.save_data(dataset)
                if self.last_filename == '':
                    self.data_saved_counter = 0
                else:
                    self.data_saved_counter += 1
            
            else:
                filename = self.last_filename.replace('.npy','') + '_' + str(self.data_saved_counter) + '.npy'
                file.save_data(dataset,filename)
                self.data_saved_counter += 1


        
        
class Recorder(object):
    def __init__(self,settings):
        self.settings = settings
        self.osc_time_axis=np.arange(0,(self.settings.num_chunks*self.settings.chunk_size)/self.settings.fs,1/self.settings.fs)
        self.osc_freq_axis=np.fft.rfftfreq(len(self.osc_time_axis),1/self.settings.fs)
        self.osc_time_data=np.zeros(shape=((self.settings.num_chunks*self.settings.chunk_size),self.settings.channels))  
        self.osc_time_data_windowed=np.zeros_like(self.osc_time_data)
        if (self.settings.num_chunks*self.settings.chunk_size)%2==0:
            self.osc_freq_data=np.zeros(shape=(int(((self.settings.num_chunks*self.settings.chunk_size)/2)+1),self.settings.channels))
        else:
            self.osc_freq_data=np.zeros(shape=(int((self.settings.num_chunks*self.settings.chunk_size+1)/2),self.settings.channels))
        
        #rounds up the number of chunks needed in the pretrig array    
        self.stored_num_chunks=int(np.ceil((self.settings.stored_time*self.settings.fs)/self.settings.chunk_size))
        #the +2 is to allow for the updating process on either side
        self.stored_time_data=np.zeros(shape=((self.stored_num_chunks*self.settings.chunk_size)+2,self.settings.channels))
        self.stored_time_data_windowed=np.zeros_like(self.stored_time_data)
        #note the +2s to match up the length of stored_num_chunks
        #formula used from the np.fft.rfft documentation
        if (self.stored_num_chunks*self.settings.chunk_size)%2==0:
            self.stored_freq_data=np.zeros(shape=(int(((
                    self.stored_num_chunks*self.settings.chunk_size+2)/2)+1),self.settings.channels))
        else:
            self.stored_freq_data=np.zeros(shape=(int((
                    self.stored_num_chunks*self.settings.chunk_size+2+1)/2),self.settings.channels))
            
            
    def __call__(self, in_data, frame_count, time_info, status):
        '''
        Obtains data from the audio stream.
        '''
        self.osc_data_chunk = (np.frombuffer(in_data, dtype=eval('np.int'+str(self.settings.nbits)))/2**(self.settings.nbits-1))
        self.osc_data_chunk=np.reshape(self.osc_data_chunk,[self.settings.chunk_size,self.settings.channels])
        for i in range(self.settings.channels):
            self.osc_time_data[:-(self.settings.chunk_size),i] = self.osc_time_data[self.settings.chunk_size:,i]
            self.osc_time_data[-(self.settings.chunk_size):,i] = self.osc_data_chunk[:,i]
            self.stored_time_data[:-(self.settings.chunk_size),i] = self.stored_time_data[self.settings.chunk_size:,i]
            self.stored_time_data[-(self.settings.chunk_size):,i] = self.osc_data_chunk[:,i]
        return (in_data, pyaudio.paContinue)
    
    
    def init_pyaudio(self,settings,_input_=True,_output_=False):
        '''
        Initialises an audio stream. Gives the user a choice of which device to access.
        '''
        
        self.audio = pyaudio.PyAudio()
        
        if settings.device_index == None:
    
            device_count = self.audio.get_device_count()
            print ('Number of devices available is: %i' %device_count)
            print ('')
            print('Devices available, by index:')
            print ('')
            for i in range(device_count):
                device = self.audio.get_device_info_by_index(i)
                print(device['index'], device['name'])
            print ('')
            default_device = self.audio.get_default_input_device_info()
            print('Default device is: %i %s'
                  %(default_device['index'],default_device['name']))
            print ('')
            settings.device_index=int(input('Insert index of required device:'))
            
        settings.device_name = self.audio.get_device_info_by_index(settings.device_index)['name']
        settings.device_full_info = self.audio.get_device_info_by_index(settings.device_index)
        
        print(settings.device_full_info['index'], settings.device_full_info['name'])
        print("Selected device: %i : %s" %(settings.device_index,settings.device_name))    
        
        self.audio_stream = self.audio.open(format=settings.format,
                                        channels=settings.channels,
                                        rate=settings.fs,
                                        input=_input_,
                                        output=_output_,
                                        frames_per_buffer=settings.chunk_size,
                                        input_device_index=settings.device_index,
                                        stream_callback=self.__call__)  
        self.audio_stream.start_stream()
        
#        return (audio_stream,audio)
        
    
    def end_pyaudio(self):
        '''
        Closes an audio stream.
        '''
        if not self.audio_stream.is_stopped():
            self.audio_stream.stop_stream()
        else: 
            pass
        self.audio_stream.close()
        self.audio.terminate()
        
        

    

     

    
class KeyPressWindow(pg.GraphicsWindow):
    '''
    A subclass of pyQtGraph GraphicsWindow that emits a signal when a key is pressed.
    
    '''
    sigKeyPress = QtCore.pyqtSignal(object)
    

    def __init__(self, rec, *args, **kwargs):
        '''
        Re-implmented from parent.
        '''
        super().__init__(*args, **kwargs)
        self.rec = rec

    def keyPressEvent(self, evt):
        '''
        Emits a signal upon a key press 
        '''
        self.scene().keyPressEvent(evt)
        self.sigKeyPress.emit(evt)
        
    def closeEvent(self,event):
        '''
        Stops QTimer,exits QApplication and closes the audio stream when the user exits the Oscilloscope window.
        '''
        Oscilloscope.timer.stop()
        self.close()
        self.rec.end_pyaudio()
        


        


   

         


# -*- coding: utf-8 -*-
"""
Created on Mon Aug 13 15:37:24 2018

@author: ae407
"""
import numpy as np
import scipy as sp
import peakutils
import pydvma as dvma
        

DEFAULT_SETTINGS = dvma.MySettings(channels=1,fs=8000,stored_time=2,viewed_time=2,device_index=1)


class Beam(object):
    ''' Defines Beam data for m1 lab'''
    
    def __init__(self,*,L,b,d,M):
        """
        Args:
            L: length (mm)
            b: width (mm)
            d: thickness (mm)
            M: mass (grams)
            
        Note units are in millimeters or grams

        """
        self.L = L/1e3
        self.b = b/1e3
        self.d = d/1e3
        self.M = M/1e3
        self.rho = self.M/(self.L*self.b*self.d)
        self.I = (self.b*(self.d**3))/12
        self.A = self.b*self.d
        self.alphaL = np.array([4.730, 7.853, 10.996, 14.137])
        

def E_calculator(*,beam,frequency,mode):
    '''
    Args:
        beam (class): Beam data object
        frequency: frequency of mode in Hz.
        mode (int): mode number

    Returns:
        E: Calculation of Young's Modulus
    '''
    rho = beam.rho
    
    if type(mode)!=int:
        print('Insert an integer for mode')
        return
    elif mode < 1 or mode > 4:
        print('Invalid mode number. Please give a number between 1 and 4')
        return
    
    alpha = beam.alphaL[mode-1] / beam.L
    
    E = ((2*np.pi*frequency)**2)*((1/alpha)**4)*beam.rho*beam.A/beam.I
    print('The Young\'s Modulus calculated from Mode %i is %.2f GPa' %(mode,E*1e-9))  
    print('Density = %.2f kgm^-3' % beam.rho)
    return E,rho
        
    
    
#####some data####
"""
beam_data=m1.beam(200,25,1.37,16)
f1=153.45
f2=424.22
f3=833.09
f4=1387.50
"""


def find_peaks(freq_data,channel=0,threshold=0,freq_range=None):
    '''
    Returns a list of the n most prominent peaks in plot.
    
    Args:
        freq_data:
        freq_range:
    '''
    if freq_range == None:
        ### use all data
        freq_range = freq_data.freq_axis[[0,-1]]
        
    elif freq_range.__class__.__name__ == 'PlotData':
        threshold=freq_range.ax.get_ybound()
        threshold=threshold[0]
        freq_range=freq_range.ax.get_xbound()
        
        

    s1 = freq_data.freq_axis >= freq_range[0]
    s2 = freq_data.freq_axis <= freq_range[1]
    selection = s1 & s2
    
    faxis = freq_data.freq_axis[selection]
    fdata = freq_data.freq_data[selection,:]
    fdata = fdata[:,channel]
    
    peaks = sp.signal.find_peaks(np.abs(fdata))
    peaks = peaks[0]
    
    df = freq_data.freq_axis[1]
    peaks = peakutils.indexes(20*np.log10(np.abs(fdata)), thres=threshold, min_dist=np.round(10/df),thres_abs=True)

    peak_frequencies = faxis[peaks]
    peak_amplitudes= np.abs(fdata[peaks])
    

    
    if len(peak_frequencies) < 10:    
        print('The most significant peaks above %.2f dB in the specified range are at the following frequencies:' %(threshold))
        freq_list = ''
        for f in peak_frequencies:
            freq_list = freq_list + '%.4g'%f + '  '
        print(freq_list)
    else:
        print('%i peaks have been found. The first 10 are:' % len(peak_frequencies))
        freq_list = ''
        for f in peak_frequencies[0:10]:
            freq_list = freq_list + '%.4g'%f + '  '
        print('To reduce the number of peaks, use a higher threshold. If using a plot to specifiy the threshold then use the zoom to view only the peaks and not the noisy data underneath.')
        
    return peak_frequencies, peak_amplitudes

